(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define('bpmn-flow-engine', ['exports'], factory) :
    (factory((global['bpmn-flow-engine'] = {})));
}(this, (function (exports) { 'use strict';

    /*! *****************************************************************************
    Copyright (c) Microsoft Corporation. All rights reserved.
    Licensed under the Apache License, Version 2.0 (the "License"); you may not use
    this file except in compliance with the License. You may obtain a copy of the
    License at http://www.apache.org/licenses/LICENSE-2.0

    THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
    KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
    WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
    MERCHANTABLITY OR NON-INFRINGEMENT.

    See the Apache Version 2.0 License for specific language governing permissions
    and limitations under the License.
    ***************************************************************************** */
    function __values(o) {
        var m = typeof Symbol === "function" && o[Symbol.iterator], i = 0;
        if (m)
            return m.call(o);
        return {
            next: function () {
                if (o && i >= o.length)
                    o = void 0;
                return { value: o && o[i++], done: !o };
            }
        };
    }
    function __read(o, n) {
        var m = typeof Symbol === "function" && o[Symbol.iterator];
        if (!m)
            return o;
        var i = m.call(o), r, ar = [], e;
        try {
            while ((n === void 0 || n-- > 0) && !(r = i.next()).done)
                ar.push(r.value);
        }
        catch (error) {
            e = { error: error };
        }
        finally {
            try {
                if (r && !r.done && (m = i["return"]))
                    m.call(i);
            }
            finally {
                if (e)
                    throw e.error;
            }
        }
        return ar;
    }
    function __spread() {
        for (var ar = [], i = 0; i < arguments.length; i++)
            ar = ar.concat(__read(arguments[i]));
        return ar;
    }

    /**
     * @fileoverview added by tsickle
     * Generated from: lib/models/flow-item.ts
     * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
     */
    /** @enum {string} */
    var FlowItemType = {
        Event: "event",
        Task: "task",
        ExclusiveGateway: "exclusivegateway",
        Sequence: "sequence",
        Process: "process",
        SubProcess: "subprocess",
    };

    /**
     * @fileoverview added by tsickle
     * Generated from: lib/models/flow-execution.ts
     * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
     */
    /** @enum {string} */
    var ExecutionState = {
        Idle: "idle",
        Running: "running",
        Stopped: "stopped",
        Finished: "finished",
    };

    /**
     * @fileoverview added by tsickle
     * Generated from: lib/models/flow-event.ts
     * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
     */
    /** @enum {string} */
    var FlowEventActionType = {
        Jump: "jump",
        SaveAndJump: "saveAndJump",
        Finish: "finish",
        FinishSubProcess: "finishsubprocess",
        GeneratePending: "generatepending",
    };
    /** @enum {string} */
    var FlowEventType = {
        Start: "start",
        End: "end",
    };

    /**
     * @fileoverview added by tsickle
     * Generated from: lib/models/flow-engine.ts
     * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
     */
    /**
     * Create a new FlowEngine instance to control a process flow.
     */
    var /**
     * Create a new FlowEngine instance to control a process flow.
     */ FlowEngine = /** @class */ (function () {
        function FlowEngine(schema, configuration) {
            var e_1, _a;
            // public readonly items: FlowItem[];
            this.items = new Map();
            this.execution = {
                state: ExecutionState.Idle,
                current: undefined,
                startNode: undefined
            };
            this.traverseHelper = [];
            this.schema = schema;
            /** @type {?} */
            var allItems = __spread(schema.tasks, schema.sequenceFlow, schema.exclusiveGateways, schema.startEvents, schema.endEvents, schema.subProcess);
            try {
                for (var allItems_1 = __values(allItems), allItems_1_1 = allItems_1.next(); !allItems_1_1.done; allItems_1_1 = allItems_1.next()) {
                    var item = allItems_1_1.value;
                    this.items.set(item.id, item);
                }
            }
            catch (e_1_1) {
                e_1 = { error: e_1_1 };
            }
            finally {
                try {
                    if (allItems_1_1 && !allItems_1_1.done && (_a = allItems_1.return))
                        _a.call(allItems_1);
                }
                finally {
                    if (e_1)
                        throw e_1.error;
                }
            }
            this.configuration = configuration;
        }
        /**
         * @return {?}
         */
        FlowEngine.prototype.getState = /**
         * @return {?}
         */
            function () {
                return this.execution.state;
            };
        /**
         * @return {?}
         */
        FlowEngine.prototype.start = /**
         * @return {?}
         */
            function () {
                var _this = this;
                this.execution.state = ExecutionState.Running;
                if (this.configuration.startItemId) {
                    this.execution.current = this.items.get(this.configuration.startItemId);
                    if (this.configuration.itemCallback)
                        this.configuration.itemCallback(this.execution.current, null);
                }
                /** @type {?} */
                var startId = (this.configuration.startItemId) ? this.configuration.startItemId : undefined;
                if (!startId && this.schema.startEvents.length === 1) {
                    startId = this.schema.startEvents[0].id;
                }
                if (!startId) {
                    this.execution.state = ExecutionState.Stopped;
                    throw (new Error("\n            Start event not provided <=======\n            Your schema must have either one start event or specify the startItemId at the configuration object."));
                }
                /** @type {?} */
                var startNode = this.items.get(startId);
                if (startNode && startNode.type !== FlowItemType.Event
                    && (( /** @type {?} */(startNode))).eventType !== FlowEventType.Start) {
                    throw (new Error("\n            Start item must be a FlowItem type: event. <=======\n            You provided a startItemId that's associated with an event. Ensure the startItemId\n            is an event and it has the EventType = start."));
                }
                this.execution.startNode = startNode;
                this.execution.current = startNode;
                /** @type {?} */
                var disconnected = this.disconnectedNodes(this.execution.startNode);
                disconnected.map(( /**
                 * @param {?} disconnectedItem
                 * @return {?}
                 */function (disconnectedItem) {
                    disconnectedItem.discarded = true;
                    disconnectedItem.visited = false;
                    if (_this.configuration.discardedCallback)
                        _this.configuration.discardedCallback(disconnectedItem);
                }));
            };
        /**
         * @param {?} itemId
         * @return {?}
         */
        FlowEngine.prototype.resume = /**
         * @param {?} itemId
         * @return {?}
         */
            function (itemId) {
                this.execution.state = ExecutionState.Running;
                this.execution.current = this.items.get(itemId);
                console.log(this.execution.current);
                if (this.configuration.itemCallback)
                    this.configuration.itemCallback(this.execution.current, null);
            };
        /**
         * @return {?}
         */
        FlowEngine.prototype.nextStep = /**
         * @return {?}
         */
            function () {
                var _this = this;
                var e_2, _a;
                // Next line returns string or array of strings depending on the flow.
                /** @type {?} */
                var outgoing = this.execution.current.outgoing;
                this.execution.current.visited = true;
                this.execution.current.discarded = false;
                // TODO: Obtener objeto modificadores del diagrama.
                // Los modificadores ejecutan un cambio de valor de otro elemento desde
                // la pregunta actual. Por ejemplo: AL pasar por el FlowItem actual,
                // vaya a la pregunta XXX y cambie su valor a undefined.
                // if (this.execution.current.id === 'tiene_techo') {
                //     this.executeModifiers(this.execution.current);
                // }
                /** @type {?} */
                var nextItem;
                if (outgoing) {
                    if (Array.isArray(outgoing)) {
                        // Convert current item in ExclusiveGateway
                        /** @type {?} */
                        var currentGateway_1 = ( /** @type {?} */(this.execution.current));
                        /** @type {?} */
                        var closureItem = this.schema.exclusiveGateways.find(( /**
                         * @param {?} i
                         * @return {?}
                         */function (i) { return i.id === currentGateway_1.closingElementId; }));
                        /** @type {?} */
                        var nextItemId = this.configuration.gatewayCallback(currentGateway_1);
                        nextItem = this.items.get(nextItemId);
                        this.mapFlowItems(nextItem, closureItem, ( /**
                         * @param {?} item
                         * @return {?}
                         */function (item) {
                            item.discarded = false;
                            if (_this.configuration.activatedCallback)
                                _this.configuration.activatedCallback(item);
                        }));
                        try {
                            // Clean other gateway data.
                            for (var outgoing_1 = __values(outgoing), outgoing_1_1 = outgoing_1.next(); !outgoing_1_1.done; outgoing_1_1 = outgoing_1.next()) {
                                var itemId = outgoing_1_1.value;
                                // Clean only those who does not belong to the next path
                                if (itemId !== nextItem.id) {
                                    /** @type {?} */
                                    var clearItem = this.items.get(itemId);
                                    this.mapFlowItems(clearItem, closureItem, ( /**
                                     * @param {?} item
                                     * @return {?}
                                     */function (item) {
                                        item.visited = false;
                                        item.discarded = true;
                                        if (_this.configuration.discardedCallback)
                                            _this.configuration.discardedCallback(item);
                                    }));
                                }
                            }
                        }
                        catch (e_2_1) {
                            e_2 = { error: e_2_1 };
                        }
                        finally {
                            try {
                                if (outgoing_1_1 && !outgoing_1_1.done && (_a = outgoing_1.return))
                                    _a.call(outgoing_1);
                            }
                            finally {
                                if (e_2)
                                    throw e_2.error;
                            }
                        }
                    }
                    else {
                        nextItem = this.items.get(outgoing);
                    }
                    /** @type {?} */
                    var prevItem = this.execution.current;
                    this.execution.current = nextItem;
                    if (nextItem.type === FlowItemType.ExclusiveGateway && this.configuration.skipGateway) {
                        return this.nextStep();
                    }
                    if (nextItem.type === FlowItemType.Sequence && this.configuration.skipSequence) {
                        return this.nextStep();
                    }
                    if (this.configuration.itemCallback)
                        this.configuration.itemCallback(nextItem, prevItem);
                    return this.execution.current;
                }
                else {
                    console.log('finished');
                    this.execution.state = ExecutionState.Finished;
                    if (this.configuration.finishedCallback)
                        this.configuration.finishedCallback();
                    return undefined;
                }
            };
        /**
         * @param {?} item
         * @return {?}
         */
        FlowEngine.prototype.executeModifiers = /**
         * @param {?} item
         * @return {?}
         */
            function (item) {
                // TODO: Exec modifiers. This function might be deleted to execute a callback instead.
            };
        /**
         * @return {?}
         */
        FlowEngine.prototype.prevStep = /**
         * @return {?}
         */
            function () {
                var _this = this;
                // Next line returns string or array of strings.
                /** @type {?} */
                var incoming = this.execution.current.incoming;
                // TODO: Obtener objeto modificadores del diagrama
                /** @type {?} */
                var prevItem;
                if (incoming) {
                    if (Array.isArray(incoming)) {
                        /** @type {?} */
                        var prevItems = incoming.map(( /**
                         * @param {?} iconmingId
                         * @return {?}
                         */function (iconmingId) { return _this.items.get(iconmingId); }));
                        prevItem = prevItems.find(( /**
                         * @param {?} i
                         * @return {?}
                         */function (i) { return (i.visited && !i.discarded); }));
                    }
                    else {
                        prevItem = this.items.get(incoming);
                    }
                    /** @type {?} */
                    var lastItem = this.execution.current;
                    this.execution.current = prevItem;
                    if ((this.configuration.skipSequence &&
                        prevItem.type === FlowItemType.Sequence) ||
                        (this.configuration.skipGateway &&
                            prevItem.type === FlowItemType.ExclusiveGateway)) {
                        return this.prevStep();
                    }
                    if (this.configuration.itemCallback)
                        this.configuration.itemCallback(this.execution.current, lastItem);
                    return this.execution.current;
                }
                else {
                    return undefined;
                }
            };
        /**
         * @param {?} from
         * @param {?} to
         * @param {?} cb
         * @return {?}
         */
        FlowEngine.prototype.mapFlowItems = /**
         * @param {?} from
         * @param {?} to
         * @param {?} cb
         * @return {?}
         */
            function (from, to, cb) {
                var e_3, _a;
                /** @type {?} */
                var current = from;
                while (current !== to) {
                    // console.log(`Executing cb for ${current.id}`);
                    cb(current);
                    if (Array.isArray(current.outgoing)) {
                        /** @type {?} */
                        var currentGateway = ( /** @type {?} */(current));
                        /** @type {?} */
                        var closureItem = ( /** @type {?} */(this.items.get(currentGateway.closingElementId)));
                        try {
                            for (var _b = __values(current.outgoing), _c = _b.next(); !_c.done; _c = _b.next()) {
                                var itemId = _c.value;
                                // Clean only those who does not belong to the next path
                                /** @type {?} */
                                var clearItem = this.items.get(itemId);
                                this.mapFlowItems(clearItem, closureItem, cb);
                            }
                        }
                        catch (e_3_1) {
                            e_3 = { error: e_3_1 };
                        }
                        finally {
                            try {
                                if (_c && !_c.done && (_a = _b.return))
                                    _a.call(_b);
                            }
                            finally {
                                if (e_3)
                                    throw e_3.error;
                            }
                        }
                        current = closureItem;
                    }
                    else {
                        if (current.outgoing) {
                            current = this.items.get(current.outgoing);
                        }
                        else {
                            // If there is no outgoing item, then the map has reached an end event, so we need to stop the loop.
                            break;
                        }
                    }
                }
            };
        /**
         * Recursive function that runs all over the graph.
         * @param node The starting node to start calculating the paths
         * @param path used for recursive purposes. Use [] as initial value.
         */
        /**
         * Recursive function that runs all over the graph.
         * @private
         * @param {?} node The starting node to start calculating the paths
         * @param {?} path used for recursive purposes. Use [] as initial value.
         * @param {?=} visited
         * @return {?}
         */
        FlowEngine.prototype.traverse = /**
         * Recursive function that runs all over the graph.
         * @private
         * @param {?} node The starting node to start calculating the paths
         * @param {?} path used for recursive purposes. Use [] as initial value.
         * @param {?=} visited
         * @return {?}
         */
            function (node, path, visited) {
                var e_4, _a;
                if (!visited) {
                    visited = new Set();
                }
                path.push(node);
                if ((node.type === FlowItemType.Event
                    && (( /** @type {?} */(node))).eventType === FlowEventType.End) || visited.has(node)) {
                    // console.log("Found Valid", path);
                    this.traverseHelper.push(this.clone(path));
                    return;
                }
                visited.add(node);
                if (Array.isArray(node.outgoing)) {
                    try {
                        for (var _b = __values(node.outgoing), _c = _b.next(); !_c.done; _c = _b.next()) {
                            var next = _c.value;
                            /** @type {?} */
                            var newPath = this.clone(path);
                            this.traverse(this.items.get(next), newPath, visited);
                        }
                    }
                    catch (e_4_1) {
                        e_4 = { error: e_4_1 };
                    }
                    finally {
                        try {
                            if (_c && !_c.done && (_a = _b.return))
                                _a.call(_b);
                        }
                        finally {
                            if (e_4)
                                throw e_4.error;
                        }
                    }
                }
                else {
                    /** @type {?} */
                    var newPath = this.clone(path);
                    this.traverse(this.items.get(node.outgoing), newPath, visited);
                }
            };
        /**
         * Returns all posible paths from one node to all End Events.
         * @param node The starting node to start calculating the paths
         */
        /**
         * Returns all posible paths from one node to all End Events.
         * @param {?} node The starting node to start calculating the paths
         * @return {?}
         */
        FlowEngine.prototype.getPaths = /**
         * Returns all posible paths from one node to all End Events.
         * @param {?} node The starting node to start calculating the paths
         * @return {?}
         */
            function (node) {
                this.traverseHelper = [];
                this.traverse(node, []);
                return this.traverseHelper;
            };
        /**
         * Get all nodes that wont be accessible from this node.
         * @param node The starting node to begin the calculation from.
         */
        /**
         * Get all nodes that wont be accessible from this node.
         * @param {?} node The starting node to begin the calculation from.
         * @return {?}
         */
        FlowEngine.prototype.disconnectedNodes = /**
         * Get all nodes that wont be accessible from this node.
         * @param {?} node The starting node to begin the calculation from.
         * @return {?}
         */
            function (node) {
                var e_5, _a;
                // Get list of lists of FlowItem (or list of paths with FlowItem inside)
                /** @type {?} */
                var validPaths = this.getPaths(node);
                // Merge all flow items inside one array
                /** @type {?} */
                var merged = [].concat.apply([], validPaths);
                /** @type {?} */
                var connectedNodes = new Set();
                try {
                    for (var merged_1 = __values(merged), merged_1_1 = merged_1.next(); !merged_1_1.done; merged_1_1 = merged_1.next()) {
                        var item = merged_1_1.value;
                        if (!connectedNodes.has(item.id)) {
                            connectedNodes.add(item.id);
                        }
                    }
                }
                catch (e_5_1) {
                    e_5 = { error: e_5_1 };
                }
                finally {
                    try {
                        if (merged_1_1 && !merged_1_1.done && (_a = merged_1.return))
                            _a.call(merged_1);
                    }
                    finally {
                        if (e_5)
                            throw e_5.error;
                    }
                }
                return Array.from(this.items.values()).filter(( /**
                 * @param {?} x
                 * @return {?}
                 */function (x) { return !connectedNodes.has(x.id); }));
            };
        /**
         * @param {?} A
         * @return {?}
         */
        FlowEngine.prototype.clone = /**
         * @param {?} A
         * @return {?}
         */
            function (A) {
                return JSON.parse(JSON.stringify(A));
            };
        return FlowEngine;
    }());

    /**
     * @fileoverview added by tsickle
     * Generated from: lib/models/interview-answer.ts
     * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
     */
    /** @enum {string} */
    var QuestionType = {
        UNIQUE_SELECTION_QUESTION: "UNIQUE_SELECTION_QUESTION",
        YN_QUESTION: "YN_QUESTION",
        INTEGER_GRID_QUESTION: "INTEGER_GRID_QUESTION",
        PLAIN_TEXT_QUESTION: "PLAIN_TEXT_QUESTION",
        SCALED_QUESTION: "SCALED_QUESTION",
        BOOLEAN_QUESTION: "BOOLEAN_QUESTION",
        SELECTION_QUESTION: "SELECTION_QUESTION",
        DATE_PICKER_QUESTION: "DATE_PICKER_QUESTION",
        INPUT_INLINE_QUESTION: "INPUT_INLINE_QUESTION",
        // Multi-instance questions
        HOME_LIST_MULTI_INSTANCE: "HOME_LIST_MULTI_INSTANCE",
        PEOPLE_LIST_MULTI_INSTANCE: "PEOPLE_LIST_MULTI_INSTANCE",
    };

    /**
     * @fileoverview added by tsickle
     * Generated from: lib/models/interview-context-item.ts
     * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
     */
    /** @enum {string} */
    var ContextItemStatus = {
        VALID: "valid",
        INCONSISTENT: "inconsistent",
        PENDING: "pending",
        DISCARDED: "discarded",
        BLOCKED: "blocked",
    };

    /**
     * @fileoverview added by tsickle
     * Generated from: public-api.ts
     * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
     */

    /**
     * @fileoverview added by tsickle
     * Generated from: bpmn-flow-engine.ts
     * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
     */

    exports.FlowEngine = FlowEngine;
    exports.FlowEventActionType = FlowEventActionType;
    exports.FlowEventType = FlowEventType;
    exports.ExecutionState = ExecutionState;
    exports.FlowItemType = FlowItemType;
    exports.QuestionType = QuestionType;
    exports.ContextItemStatus = ContextItemStatus;

    Object.defineProperty(exports, '__esModule', { value: true });

})));

//# sourceMappingURL=bpmn-flow-engine.umd.js.map