/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/flow-item.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @enum {string} */
const FlowItemType = {
    Event: "event",
    Task: "task",
    ExclusiveGateway: "exclusivegateway",
    Sequence: "sequence",
    Process: "process",
    SubProcess: "subprocess",
};

/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/flow-execution.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @enum {string} */
const ExecutionState = {
    Idle: "idle",
    Running: "running",
    Stopped: "stopped",
    Finished: "finished",
};

/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/flow-event.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @enum {string} */
const FlowEventActionType = {
    Jump: "jump",
    SaveAndJump: "saveAndJump",
    Finish: "finish",
    FinishSubProcess: "finishsubprocess",
    GeneratePending: "generatepending",
};
/** @enum {string} */
const FlowEventType = {
    Start: "start",
    End: "end",
};

/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/flow-engine.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Create a new FlowEngine instance to control a process flow.
 */
class FlowEngine {
    /**
     * @param {?} schema
     * @param {?} configuration
     */
    constructor(schema, configuration) {
        // public readonly items: FlowItem[];
        this.items = new Map();
        this.execution = {
            state: ExecutionState.Idle,
            current: undefined,
            startNode: undefined
        };
        this.traverseHelper = [];
        this.schema = schema;
        /** @type {?} */
        const allItems = [
            ...schema.tasks,
            ...schema.sequenceFlow,
            ...schema.exclusiveGateways,
            ...schema.startEvents,
            ...schema.endEvents,
            ...schema.subProcess
        ];
        for (const item of allItems) {
            this.items.set(item.id, item);
        }
        this.configuration = configuration;
    }
    /**
     * @return {?}
     */
    getState() {
        return this.execution.state;
    }
    /**
     * @return {?}
     */
    start() {
        this.execution.state = ExecutionState.Running;
        if (this.configuration.startItemId) {
            this.execution.current = this.items.get(this.configuration.startItemId);
            if (this.configuration.itemCallback)
                this.configuration.itemCallback(this.execution.current, null);
        }
        /** @type {?} */
        let startId = (this.configuration.startItemId) ? this.configuration.startItemId : undefined;
        if (!startId && this.schema.startEvents.length === 1) {
            startId = this.schema.startEvents[0].id;
        }
        if (!startId) {
            this.execution.state = ExecutionState.Stopped;
            throw (new Error(`
            Start event not provided <=======
            Your schema must have either one start event or specify the startItemId at the configuration object.`));
        }
        /** @type {?} */
        const startNode = this.items.get(startId);
        if (startNode && startNode.type !== FlowItemType.Event
            && ((/** @type {?} */ (startNode))).eventType !== FlowEventType.Start) {
            throw (new Error(`
            Start item must be a FlowItem type: event. <=======
            You provided a startItemId that's associated with an event. Ensure the startItemId
            is an event and it has the EventType = start.`));
        }
        this.execution.startNode = startNode;
        this.execution.current = startNode;
        /** @type {?} */
        const disconnected = this.disconnectedNodes(this.execution.startNode);
        disconnected.map((/**
         * @param {?} disconnectedItem
         * @return {?}
         */
        disconnectedItem => {
            disconnectedItem.discarded = true;
            disconnectedItem.visited = false;
            if (this.configuration.discardedCallback)
                this.configuration.discardedCallback(disconnectedItem);
        }));
    }
    /**
     * @param {?} itemId
     * @return {?}
     */
    resume(itemId) {
        this.execution.state = ExecutionState.Running;
        this.execution.current = this.items.get(itemId);
        console.log(this.execution.current);
        if (this.configuration.itemCallback)
            this.configuration.itemCallback(this.execution.current, null);
    }
    /**
     * @return {?}
     */
    nextStep() {
        // Next line returns string or array of strings depending on the flow.
        /** @type {?} */
        const outgoing = this.execution.current.outgoing;
        this.execution.current.visited = true;
        this.execution.current.discarded = false;
        // TODO: Obtener objeto modificadores del diagrama.
        // Los modificadores ejecutan un cambio de valor de otro elemento desde
        // la pregunta actual. Por ejemplo: AL pasar por el FlowItem actual,
        // vaya a la pregunta XXX y cambie su valor a undefined.
        // if (this.execution.current.id === 'tiene_techo') {
        //     this.executeModifiers(this.execution.current);
        // }
        /** @type {?} */
        let nextItem;
        if (outgoing) {
            if (Array.isArray(outgoing)) {
                // Convert current item in ExclusiveGateway
                /** @type {?} */
                const currentGateway = (/** @type {?} */ (this.execution.current));
                /** @type {?} */
                const closureItem = this.schema.exclusiveGateways.find((/**
                 * @param {?} i
                 * @return {?}
                 */
                i => i.id === currentGateway.closingElementId));
                /** @type {?} */
                const nextItemId = this.configuration.gatewayCallback(currentGateway);
                nextItem = this.items.get(nextItemId);
                this.mapFlowItems(nextItem, closureItem, (/**
                 * @param {?} item
                 * @return {?}
                 */
                (item) => {
                    item.discarded = false;
                    if (this.configuration.activatedCallback)
                        this.configuration.activatedCallback(item);
                }));
                // Clean other gateway data.
                for (const itemId of outgoing) {
                    // Clean only those who does not belong to the next path
                    if (itemId !== nextItem.id) {
                        /** @type {?} */
                        const clearItem = this.items.get(itemId);
                        this.mapFlowItems(clearItem, closureItem, (/**
                         * @param {?} item
                         * @return {?}
                         */
                        (item) => {
                            item.visited = false;
                            item.discarded = true;
                            if (this.configuration.discardedCallback)
                                this.configuration.discardedCallback(item);
                        }));
                    }
                }
            }
            else {
                nextItem = this.items.get(outgoing);
            }
            /** @type {?} */
            const prevItem = this.execution.current;
            this.execution.current = nextItem;
            if (nextItem.type === FlowItemType.ExclusiveGateway && this.configuration.skipGateway) {
                return this.nextStep();
            }
            if (nextItem.type === FlowItemType.Sequence && this.configuration.skipSequence) {
                return this.nextStep();
            }
            if (this.configuration.itemCallback)
                this.configuration.itemCallback(nextItem, prevItem);
            return this.execution.current;
        }
        else {
            console.log('finished');
            this.execution.state = ExecutionState.Finished;
            if (this.configuration.finishedCallback)
                this.configuration.finishedCallback();
            return undefined;
        }
    }
    /**
     * @param {?} item
     * @return {?}
     */
    executeModifiers(item) {
        // TODO: Exec modifiers. This function might be deleted to execute a callback instead.
    }
    /**
     * @return {?}
     */
    prevStep() {
        // Next line returns string or array of strings.
        /** @type {?} */
        const incoming = this.execution.current.incoming;
        // TODO: Obtener objeto modificadores del diagrama
        /** @type {?} */
        let prevItem;
        if (incoming) {
            if (Array.isArray(incoming)) {
                /** @type {?} */
                const prevItems = incoming.map((/**
                 * @param {?} iconmingId
                 * @return {?}
                 */
                iconmingId => this.items.get(iconmingId)));
                prevItem = prevItems.find((/**
                 * @param {?} i
                 * @return {?}
                 */
                i => (i.visited && !i.discarded)));
            }
            else {
                prevItem = this.items.get(incoming);
            }
            /** @type {?} */
            const lastItem = this.execution.current;
            this.execution.current = prevItem;
            if ((this.configuration.skipSequence &&
                prevItem.type === FlowItemType.Sequence) ||
                (this.configuration.skipGateway &&
                    prevItem.type === FlowItemType.ExclusiveGateway)) {
                return this.prevStep();
            }
            if (this.configuration.itemCallback)
                this.configuration.itemCallback(this.execution.current, lastItem);
            return this.execution.current;
        }
        else {
            return undefined;
        }
    }
    /**
     * @param {?} from
     * @param {?} to
     * @param {?} cb
     * @return {?}
     */
    mapFlowItems(from, to, cb) {
        /** @type {?} */
        let current = from;
        while (current !== to) {
            // console.log(`Executing cb for ${current.id}`);
            cb(current);
            if (Array.isArray(current.outgoing)) {
                /** @type {?} */
                const currentGateway = (/** @type {?} */ (current));
                /** @type {?} */
                const closureItem = (/** @type {?} */ (this.items.get(currentGateway.closingElementId)));
                for (const itemId of current.outgoing) {
                    // Clean only those who does not belong to the next path
                    /** @type {?} */
                    const clearItem = this.items.get(itemId);
                    this.mapFlowItems(clearItem, closureItem, cb);
                }
                current = closureItem;
            }
            else {
                if (current.outgoing) {
                    current = this.items.get(current.outgoing);
                }
                else {
                    // If there is no outgoing item, then the map has reached an end event, so we need to stop the loop.
                    break;
                }
            }
        }
    }
    /**
     * Recursive function that runs all over the graph.
     * @private
     * @param {?} node The starting node to start calculating the paths
     * @param {?} path used for recursive purposes. Use [] as initial value.
     * @param {?=} visited
     * @return {?}
     */
    traverse(node, path, visited) {
        if (!visited) {
            visited = new Set();
        }
        path.push(node);
        if ((node.type === FlowItemType.Event
            && ((/** @type {?} */ (node))).eventType === FlowEventType.End) || visited.has(node)) {
            // console.log("Found Valid", path);
            this.traverseHelper.push(this.clone(path));
            return;
        }
        visited.add(node);
        if (Array.isArray(node.outgoing)) {
            for (const next of node.outgoing) {
                /** @type {?} */
                const newPath = this.clone(path);
                this.traverse(this.items.get(next), newPath, visited);
            }
        }
        else {
            /** @type {?} */
            const newPath = this.clone(path);
            this.traverse(this.items.get(node.outgoing), newPath, visited);
        }
    }
    /**
     * Returns all posible paths from one node to all End Events.
     * @param {?} node The starting node to start calculating the paths
     * @return {?}
     */
    getPaths(node) {
        this.traverseHelper = [];
        this.traverse(node, []);
        return this.traverseHelper;
    }
    /**
     * Get all nodes that wont be accessible from this node.
     * @param {?} node The starting node to begin the calculation from.
     * @return {?}
     */
    disconnectedNodes(node) {
        // Get list of lists of FlowItem (or list of paths with FlowItem inside)
        /** @type {?} */
        const validPaths = this.getPaths(node);
        // Merge all flow items inside one array
        /** @type {?} */
        const merged = [].concat.apply([], validPaths);
        /** @type {?} */
        const connectedNodes = new Set();
        for (const item of merged) {
            if (!connectedNodes.has(item.id)) {
                connectedNodes.add(item.id);
            }
        }
        return Array.from(this.items.values()).filter((/**
         * @param {?} x
         * @return {?}
         */
        x => !connectedNodes.has(x.id)));
    }
    /**
     * @param {?} A
     * @return {?}
     */
    clone(A) {
        return JSON.parse(JSON.stringify(A));
    }
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/flow-engine-configuration.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/flow-gateway.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/flow-schema.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/flow-sequence.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/flow-task.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/interview-answer.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @enum {string} */
const QuestionType = {
    UNIQUE_SELECTION_QUESTION: "UNIQUE_SELECTION_QUESTION",
    YN_QUESTION: "YN_QUESTION",
    INTEGER_GRID_QUESTION: "INTEGER_GRID_QUESTION",
    PLAIN_TEXT_QUESTION: "PLAIN_TEXT_QUESTION",
    SCALED_QUESTION: "SCALED_QUESTION",
    BOOLEAN_QUESTION: "BOOLEAN_QUESTION",
    SELECTION_QUESTION: "SELECTION_QUESTION",
    DATE_PICKER_QUESTION: "DATE_PICKER_QUESTION",
    INPUT_INLINE_QUESTION: "INPUT_INLINE_QUESTION",
    // Multi-instance questions
    HOME_LIST_MULTI_INSTANCE: "HOME_LIST_MULTI_INSTANCE",
    PEOPLE_LIST_MULTI_INSTANCE: "PEOPLE_LIST_MULTI_INSTANCE",
};

/**
 * @fileoverview added by tsickle
 * Generated from: lib/models/interview-context-item.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @enum {string} */
const ContextItemStatus = {
    VALID: "valid",
    INCONSISTENT: "inconsistent",
    PENDING: "pending",
    DISCARDED: "discarded",
    BLOCKED: "blocked",
};

/**
 * @fileoverview added by tsickle
 * Generated from: public-api.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: bpmn-flow-engine.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { FlowEngine, FlowEventActionType, FlowEventType, ExecutionState, FlowItemType, QuestionType, ContextItemStatus };

//# sourceMappingURL=bpmn-flow-engine.js.map