import { FlowSchema } from './flow-schema';
import { FlowItem } from './flow-item';
import { FlowExecution, ExecutionState } from './flow-execution';
import { FlowEngineConfiguration } from './flow-engine-configuration';
/**
 * Create a new FlowEngine instance to control a process flow.
 */
export declare class FlowEngine {
    readonly schema: FlowSchema;
    readonly items: Map<string, FlowItem>;
    readonly execution: FlowExecution;
    private configuration;
    private traverseHelper;
    constructor(schema: FlowSchema, configuration: FlowEngineConfiguration);
    getState(): ExecutionState;
    start(): void;
    resume(itemId: string): void;
    nextStep(): FlowItem;
    executeModifiers(item: FlowItem): void;
    prevStep(): FlowItem;
    mapFlowItems(from: FlowItem, to: FlowItem, cb: (current: FlowItem) => void): void;
    /**
     * Recursive function that runs all over the graph.
     * @param node The starting node to start calculating the paths
     * @param path used for recursive purposes. Use [] as initial value.
     */
    private traverse;
    /**
     * Returns all posible paths from one node to all End Events.
     * @param node The starting node to start calculating the paths
     */
    getPaths(node: FlowItem): FlowItem[][];
    /**
     * Get all nodes that wont be accessible from this node.
     * @param node The starting node to begin the calculation from.
     */
    disconnectedNodes(node: FlowItem): FlowItem[];
    clone(A: any): any;
}
